/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file debug.cpp
 *
 * Implements the cdebug class. For more information, see debug.h.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "debug.h"
#include <fstream>
#include <iostream>


/*****************************************************************************
 * INITIALIZE STATIC MEMBERS
 *****************************************************************************/
const string cdebug::kmaster_setting("master_debug");
const string cdebug::klogfile_setting("debug_logfile");


/*****************************************************************************
 * METHODS
 *****************************************************************************/

// ctor
//
// the default version, uses cerr
cdebug::cdebug() : out(cerr)
{
  pthread_mutex_init(&mtx, NULL);
}

// ctor
//
// the file version, uses the filename passed in
cdebug::cdebug(const string &filename) :
  file(filename.c_str(), ios::out | ios::app), // only append!
  out(file)
{
//   assert(filename != "");
  pthread_mutex_init(&mtx, NULL);

  // wipe the file at the beginning (since debug should append to it) 
  fstream wipe(filename.c_str(), ios::out | ios::trunc);
}

cdebug::~cdebug()
{
//   *this << acquire("master_debug") << "cdebug dtor()" << release();
  pthread_mutex_destroy(&mtx);
}


// acquire manipulator (parameterized)
// 
// only acquires the debug stream if debugging is on *and* the given module
// setting is true.
// declared as a friend of the cdebug class
cdebug &operator<< (cdebug &debug, const acquire &a)
{
  if (!debug.is_on(a.module))
    return debug;  // this debug module is turned off

  pthread_mutex_lock(&debug.mtx);
  debug.module = a.module;

  debug.out << a.module << ": ";

  // we don't release mutex since this module owns it now
  return debug;
}
  


// release manipulator
// 
// declared as a friend of the cdebug class
cdebug &operator<< (cdebug &debug, const release &r)
{
  if (debug.module == "")
    return debug;

  debug.out << endl;

  // make debug available to other modules
  debug.module = "";
  pthread_mutex_unlock(&debug.mtx);

  return debug;
}



/*****************************************************************************
 * TEST MAIN
 *****************************************************************************/

/*
int main(int argc, char **argv)
{
  csettings::instance().set(cdebug::kmaster_setting, true);
//   csettings::instance().set(cdebug::klogfile_setting, "debug_test.log");

  cdebug::instance() << acquire() << "asdf ";
  cdebug::instance() << "this should print ";
  cdebug::instance() << 2.34 << " " << 3;

  double asdf = 2.999999999999999999999999999;
  cdebug::instance() << asdf;

  cdebug::instance() << " but this (below) shouldn't " << release();

  csettings::instance().set(cdebug::kmaster_setting, false);

  cdebug::instance() << acquire() << "you should not see this!!!" << release();


  csettings::instance().set(cdebug::kmaster_setting, true);

  cdebug::instance() << acquire() << "but this should!!!" << release();


  cdebug::instance() << acquire("debug_test") << "this should only print if "
                     << "the debug_test setting in the .p4 file is true."
                     << release();
}
*/
