/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file lka.cpp
 *
 * This implements the lka class.  For more information, see lka.h.
 * 
 * TODO:
 * - set new handler
 * - have lookup return const_iterator some of the time
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "receiveq.h"
#include "lka.h"
#include "conn_mgr.h"
#include "settings.h"
#include "debug.h"
#include "p4error.h"

#ifdef _WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#include <arpa/inet.h>  // for inet_addr
#endif

#include <cassert>
#include <semaphore.h>
#include <cstdlib>


extern string glocal_user;

clka::clka() : entries(lka_less()) 
{
  pthread_rwlock_init(&rwlock, NULL);

  // add the local user
  struct sockaddr_in sin;
  get_local_addr(sin);

  add(csettings::instance().get<string>("local_username").c_str(),
      sin.sin_addr.s_addr, sin.sin_port, 0, 0);
}

clka::~clka()
{
  cdebug::instance() << acquire("lka_debug") << "dtor()" << release();
  pthread_rwlock_destroy(&rwlock);
}

uint clka::wipe()
{
  cdebug::instance() << acquire("lka_debug") << "wipe()" << release();

  pthread_rwlock_wrlock(&rwlock);  // get write lock

  set<lka_pair *, lka_less>::reverse_iterator entry_it;

  for (entry_it = entries.rbegin(); entry_it != entries.rend(); entry_it++) {
    delete[] (*entry_it)->second;
    delete (*entry_it);
  }

  entries = set<lka_pair*, lka_less>();
  plugin_lists = map<ushort, vector<lka_pair*> >();
  
  pthread_rwlock_unlock(&rwlock);  // release write lock

  cdebug::instance() << acquire("lka_debug") << "wiped" << release();
  return 0;
}

// See package to see how the ack is structured.
// don't need to lock. add() locks for us.
uint clka::populate(p4_lka_ack *ack)
{
  cdebug::instance() << acquire("lka_debug") << "populate()" << release();

  ulong num_entries = ack->num_entries;

  lka_entry *new_entries = reinterpret_cast<lka_entry*>
    (reinterpret_cast<char*>(ack) + sizeof(p4_lka_ack));

  plugin_info *plugins = reinterpret_cast<plugin_info*>
    (new_entries + num_entries);

  // don't need to acquire writers lock, because add() locks for us
  for (ulong i = 0; i < num_entries; i++) {
    add(new_entries[i].username, new_entries[i].addr, new_entries[i].port,
        new_entries[i].num_plugins, plugins);
    plugins += new_entries[i].num_plugins;
  }

  return kerr_success;
}

/**
 * package
 * 
 * Returns a chunk of memory that has the following format:
 * ----------------------------------------------------------------------------
 * | p4_lka_ack | lka_entry | lka_entry | ... | plugin_info | plugin_info | ...|
 * ----------------------------------------------------------------------------
 *
 * The p4_lka_ack tells how many lka_entries there are.  Each lka_entry
 * tells how many plugin_infos it has.  That means that if the first 
 * lka_entry has 3 plugins, the first three plugin_infos belong to that
 * that entry.  The second entry gets its plugin_infos starting at the
 * fourth and so on.
 */
p4_lka_ack *clka::package(ulong &package_size)
{
  pthread_rwlock_rdlock(&rwlock);

  ulong num_entries = entries.size();
  ulong num_plugins = 0;

  set<lka_pair *, lka_less>::const_iterator entry_it;

  // count the number of plugins for all users
  for (entry_it = entries.begin(); entry_it != entries.end(); entry_it++) {
    num_plugins += (*entry_it)->first.num_plugins;
  }
  
  package_size = sizeof(p4_lka_ack) + num_entries * sizeof(lka_entry) +
    sizeof(plugin_info) * num_plugins;
  char *ack_buf = new char[package_size];
  
  // set up pointers to the sections of the data
  lka_entry *entry_ptr = reinterpret_cast<lka_entry*>
    (ack_buf + sizeof(p4_lka_ack));
  plugin_info *plugin_ptr = reinterpret_cast<plugin_info*>
    (entry_ptr + num_entries);
  
  // copy all of the entries and their plugins to ack
  for (entry_it = entries.begin(); entry_it != entries.end(); entry_it++) {
    *entry_ptr = (*entry_it)->first; // copy the entry
    memcpy(plugin_ptr, (*entry_it)->second, 
         sizeof(plugin_info) * (*entry_it)->first.num_plugins);
    plugin_ptr += (*entry_it)->first.num_plugins;
    entry_ptr++;
  }
  p4_lka_ack *ack = reinterpret_cast<p4_lka_ack*>(ack_buf);
  ack->num_entries = num_entries;

  pthread_rwlock_unlock(&rwlock);

  return ack;
}
  

uint clka::add(const char *username, ulong addr, ushort port,
         ushort num_plugins, const plugin_info *plugins)
{
  struct in_addr sin = { addr };
    cdebug::instance() << acquire("lka_debug")
                       << "add(" << username  << " " << inet_ntoa(sin) << " "
                       << port << " "  << num_plugins << ")"
                       << release();

  if (!username || (num_plugins && !plugins)) {
    cdebug::instance() << acquire("lka_debug")
                       << "Couldn't add to lka "
                       << release();
    return kerr_bad_arg;
  }
  

  pthread_rwlock_wrlock(&rwlock);  // get write lock for entire function

  set<lka_pair*, lka_less>::const_iterator it = lookup(username);
  lka_pair *p;
  
  if (it == entries.end()) {
    // create new entry
    p = new lka_pair();
    strncpy(p->first.username, username, kp4_username_len);
    p->first.username[kp4_username_len - 1] = '\0';
    entries.insert(p);

  } else {
    // remove old entries in plugin_lists
    p = *it;
    for_each(p->second, p->second + p->first.num_plugins,
             plugin_list_remove(p));

    // remove old lka entry
    delete[] p->second;
  }

  // populate new entry
  p->first.addr = addr;
  p->first.port = port;
  p->first.num_plugins = num_plugins;
  
  p->second = new plugin_info[num_plugins];
  copy(plugins, plugins + num_plugins, p->second);

  // add to plugin_lists
  for_each(p->second, p->second + p->first.num_plugins, plugin_list_add(p));

  pthread_rwlock_unlock(&rwlock);  // release write lock

  return kerr_success;
}


uint clka::remove(const char *username)
{
  cdebug::instance() << acquire("lka_debug")
                     << "remove("  << username << ")"
                     << release();
  pthread_rwlock_wrlock(&rwlock);    // get write lock

  set<lka_pair*, lka_less>::iterator it = lookup(username);
  
  if (it == entries.end()) {
    pthread_rwlock_unlock(&rwlock);  // release write lock
    return kerr_lka_user_not_found;
  }
    
  lka_pair *p = *it;
    
  for_each(p->second, p->second + p->first.num_plugins,
           plugin_list_remove(p));
  entries.erase(it);
    
  pthread_rwlock_unlock(&rwlock);    // release write lock
    
  return kerr_success;
}


uint clka::get_address(const char *username, ulong &addr, ushort &port) const
{
  cdebug::instance() << acquire("lka_debug")
                     << "get_address(" << username << ")"
                     << release();
  pthread_rwlock_rdlock(&rwlock);    // get read lock

  set<lka_pair*, lka_less>::const_iterator it = lookup(username);
  
  if (it == entries.end()) {
    pthread_rwlock_unlock(&rwlock);  // release read lock
    return kerr_lka_user_not_found;
  }

  lka_pair *p = *it;

  addr = p->first.addr;
  port = p->first.port;

  pthread_rwlock_unlock(&rwlock);    // release read lock

  return kerr_success;
}


uint clka::get_num_users(ushort plugin_id, uint &num_users) const
{
  cdebug::instance() << acquire("lka_debug")
                     << "get_num_users(" << plugin_id << ")"
                     << release();

  assert(plugin_id != 0);            // error check

  pthread_rwlock_rdlock(&rwlock);    // get read lock

  map<ushort, vector<lka_pair *> >::const_iterator it =
    plugin_lists.find(plugin_id);

  if (it == plugin_lists.end())
    num_users = 0;
  else
    num_users = it->second.size();

  pthread_rwlock_unlock(&rwlock);    // release read lock

  return kerr_success;
}


uint clka::get_num_users(uint &num_users) const
{
  pthread_rwlock_rdlock(&rwlock);

  num_users = entries.size();
  pthread_rwlock_unlock(&rwlock);
  return kerr_success;
}  


uint clka::get_users(ushort plugin_id, char **users, uint num_users,
                     uint &num_written) const
{
  cdebug::instance() << acquire("lka_debug")
                     << "get_users(" << plugin_id << ")"
                     << release();

  assert(plugin_id != 0);          // error check
  if (!users)
    return kerr_bad_arg;
  for (uint i = 0; i < num_users; i++) {
    if (!users[i])
      return kerr_bad_arg;
  }

  pthread_rwlock_rdlock(&rwlock);  // get read lock

  map<ushort, vector<lka_pair *> >::const_iterator it =
    plugin_lists.find(plugin_id);


  uint num_copies = 0;             // copy users
  if (it != plugin_lists.end()) {
    // cast for mac os x
    num_copies = min(static_cast<uint>(it->second.size()), num_users);
    for (uint i = 0; i < num_copies; i++) {
      strcpy(users[i], it->second[i]->first.username);
    }
  }

  num_written = num_copies;

  pthread_rwlock_unlock(&rwlock);    // release read lock
  return kerr_success;
}


/** copy_username
 *
 * Functor for copying each user's username into the client provided array.
 * Only copies num_users usernames or less. Used by get_users.
 *
 * When done, num_copied will hold the number of usernames copied.
 */
struct copy_username
{
  copy_username(char **users, uint max_to_copy) :
    users(users), max_to_copy(max_to_copy), num_copied(0) {}

  void operator() (lka_pair *entry) {
    if (num_copied < max_to_copy)
      strcpy(users[num_copied++], entry->first.username);
  }

  char **users;
  uint max_to_copy, num_copied;
};


uint clka::get_users(char **users, uint num_users, uint &num_written) const
{
  pthread_rwlock_rdlock(&rwlock);

  uint num_entries = static_cast<uint>(entries.size());  // cast for mac os x
  copy_username functor(users, min(num_users, num_entries));
  copy_username result = for_each(entries.begin(), entries.end(), functor);

  num_written = result.num_copied;

  pthread_rwlock_unlock(&rwlock);
  return kerr_success;
}  


uint clka::get_num_plugins(const char *username, ushort &num_plugins) const
{
  cdebug::instance() << acquire("lka_debug") << "get_num_plugins("
                     << username << ")" << release();
  if (!username)
    return kerr_bad_arg;

  pthread_rwlock_rdlock(&rwlock);    // get read lock

  set<lka_pair*, lka_less>::const_iterator it = lookup(username);
  
  if (it == entries.end()) {
    pthread_rwlock_unlock(&rwlock);  // release read lock
    return kerr_lka_user_not_found;
  }

  num_plugins = (*it)->first.num_plugins;

  pthread_rwlock_unlock(&rwlock);    // release read lock

  return kerr_success;
}
  

uint clka::get_plugins(const char *username, ushort *plugin_ids, 
           ushort num_plugins, ushort &num_written) const
{
  cdebug::instance() << acquire("lka_debug") << "get_plugins("
             << username << ")" << release();
  if (!username || !plugin_ids)
    return kerr_bad_arg;
 
  pthread_rwlock_rdlock(&rwlock);    // get read lock

  set<lka_pair*, lka_less>::const_iterator it = lookup(username);
  
  if (it == entries.end()) {
    pthread_rwlock_unlock(&rwlock);  // release read lock
    return kerr_lka_user_not_found;
  }

  lka_pair *p = *it;
  uint num_copies = min(p->first.num_plugins, num_plugins);
  for (uint i = 0; i < num_copies; i++) {
    plugin_ids[i] = p->second[i].id;
  }

  pthread_rwlock_unlock(&rwlock);    // release read lock

  num_written = num_copies;

  return kerr_success;
}
// NOTE: the two lookup functions should be identical except that one
//       is const and the other is not.
// 
// don't need to acquire reader lock because lookup is a helper method
set<lka_pair*, clka::lka_less>::iterator clka::lookup(const char *username)
{
  lka_pair p;

  cdebug::instance() << acquire("lka_debug") << "lookup("
             << username << ")" << release();

  strncpy(p.first.username, username, kp4_username_len);
  p.first.username[kp4_username_len - 1] = 0;

  return entries.find(&p);

}
set<lka_pair*, clka::lka_less>::const_iterator
clka::lookup(const char *username) const
{
  lka_pair p;

  cdebug::instance() << acquire("lka_debug") << "lookup("
             << username << ")" << release();

  strncpy(p.first.username, username, kp4_username_len);
  p.first.username[kp4_username_len - 1] = 0;

  return entries.find(&p);

}


void clka::update_local_user()
{
  creceiveq &rq = creceiveq::instance();

  ushort num = rq.num_plugins();
  plugin_info *plugins = new plugin_info[num];
  
  ushort num_written;
  uint ret = rq.get_plugins(num, plugins, num_written);
  assert(ret == kerr_success);

  struct sockaddr_in sin;
  if (get_local_addr(sin) != kerr_success)
    return;

  ret = clka::instance().add(glocal_user.c_str(), sin.sin_addr.s_addr,
                 ntohs(sin.sin_port), num_written, plugins);
  assert(ret == kerr_success);
}

uint clka::random_user(int min_users, string &username) const
{
  cdebug::instance() << acquire("lka_debug") << "random_user()" <<
    release();
  pthread_rwlock_rdlock(&rwlock);

  int num_entries = entries.size();

  if (num_entries == 1 || num_entries < min_users) {
    pthread_rwlock_unlock(&rwlock);
    return kerr_no_users;
  }

  set<lka_pair*, lka_less>::const_iterator it = entries.begin();

  while (glocal_user.compare((*it)->first.username) == 0 || 
       rand() > (RAND_MAX / num_entries)) {
    it++;
    if (it == entries.end())
      it = entries.begin();
  }
  username = (*it)->first.username;

  pthread_rwlock_unlock(&rwlock);
  return kerr_success;
}

/********************************************************************************
 * TESTING CODE
 *******************************************************************************/
#ifdef LKA_MAIN
#include <cstdio>

void *writer_func(void *data)
{
    clka &lka = clka::instance();
    sem_t *done = (sem_t*)data;
       
    plugin_info plugins[3] = { { 2 }, { 4 }, { 7 } };
    char name[20];

    for (int i = 0; i < 500; i++) {
  sprintf(name, "ryan%d", i);
  lka.add(name, 54, 23, 3, plugins);
    } 
    sem_post(done);
    return NULL;
}

void *reader_func(void *data)
{
    ulong addr;
    ushort port;
    sem_t *done = (sem_t*)data;
    clka &lka = clka::instance();
    char name[20];
  
    for (int i = 0; i < 500; i++) {
    sprintf(name, "ryan%d", i);
    uint ret = lka.get_address(name, addr, port);
    if (ret == kerr_success)
      cout << name <<  " addr = " << addr << " port = " << port << endl;
    else
      cout << "couldn't find " << name << endl;
    }
    sem_post(done);
    return NULL;
}

void *remover_func(void *data)
{
    clka &lka = clka::instance();
    sem_t *done = (sem_t*)data;
    
    ulong addr;
    ushort port;

    uint ret = lka.remove("ryan0");
    ret = lka.get_address("ryan", addr, port);
    assert(ret == kerr_lka_user_not_found);
    sem_post(done);
    return NULL;
} 

void *acker_func(void *data)
{
  clka &lka = clka::instance();  
  p4_lka_ack *ack = lka.package();
  lka.wipe();
  lka.populate(ack);
  
  delete ack;

  reader_func(data);
  return NULL;
}

int main(int argc, char **argv)
{
 
  sem_t done;
  pthread_t writer, reader, remover, acker;
  
  sem_init(&done, 0, 0);
  
  pthread_create(&writer, NULL, &writer_func, (void*)&done);
  pthread_create(&reader, NULL, &reader_func, (void*)&done);
  pthread_create(&remover, NULL, &remover_func, (void*)&done);
  
  for (int i = 0; i < 3; i++) {
    sem_wait(&done);
    cout << "thread done" << endl;
  }
  
  pthread_create(&reader, NULL, &reader_func, (void*)&done);
  
  sem_wait(&done);

  pthread_create(&acker, NULL, &acker_func, (void*)&done);

  sem_wait(&done);

  cdebug::instance() << acquire("lka_debug") << "debgging statement"
             << release();

  return 0;
}

#endif // LKA_MAIN
